#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Flitt3r"
__copyright__ = f"(c) 2019 {__author__}"
__license__ = "MIT"
__version__ = "0.2.2"
__url__ = "https://gitlab.com/flitt3r/metabeyond"
